:Namespace activate ⍝ v0.3.1
⍝ Pre-packaged tool activation
⍝ 2023-09-25
⍝ 2023-11-02 kai: `Run` must go into StartupSession/ rather than SessionExtensions/
⍝ 2023-11-06 kai: One can now -reset or deactivate just one of Cider/Tatin
⍝ 2024-01-15 adam: [21074] always return result

    ⎕IO←1 ⋄ ⎕ML←1   ⍝ Set those here to avoid inheriting them from outside
    AllCmds←'Activate' 'Deactivate'

    ∇ r←List ⍝ this function usually returns 2 or more namespaces (here only 2)
      r←⎕NS¨0/¨AllCmds                             ⍝ create commands
      r.Name←AllCmds                               ⍝ their names
      r[1].Desc←'Activate or reset an experimental tool shipped with Dyalog APL'
      r[2].Desc←'Deactivate a previously activated tool'
      r.Group←'Tools' 'Tools'
     ⍝ Parsing rules for each:
      r[1].Parse←'1 -reset'
      r[2].Parse←'1'
    ∇

    ∇ r←Level Help Cmd;t
      :Select Cmd
     
      :Case 'Activate'
          r←,⊂'Activate an experimental tool shipped with Dyalog APL'
          r,←⊂'    ]',Cmd,' cider|tatin|all -update -reset'
          r,←⊂''
          r,←⊂'  -reset          revert to the version shipped with APL'
     
      :Case 'Deactivate'
          r←,⊂'Deactivate all or some previously ]Activate''d tools'
          r,←⊂'    ]',Cmd,' all|tatin|cider'
          r,←⊂''
          r,←⊂'Note that you cannot Deactivate Tatin while Cider is still activated because Cider depends on Tatin.'
     
      :EndSelect
    ∇

    ∇ r←Run(Cmd Args);tool;target;source;m;src;z;rid;s;target2;cmddir;sep;folder
      APLV←'.'⎕WG'APLVersion'
      DYALOG←2 ⎕NQ'.' 'GetEnvironment' 'DYALOG'
      :If 19>⊃2⊃'.'⎕VFI 2⊃APLV
          →0⊣r←'This user command requires Dyalog version 19.0'
      :EndIf
     
      :If WIN←'Win'≡OS←3↑1⊃APLV
          USERDIR←2⊃4070⌶⍬
      :Else
          rid←(3↑⎕D∩⍨2⊃APLV),((1+80=⎕DR'A')⊃'CU'),(1+1∊'64'⍷1⊃APLV)⊃'32' '64'
          USERDIR←(2 ⎕NQ #'GetEnvironment' 'HOME'),'/.dyalog/dyalog.',rid,'.files'
      :EndIf
     
      tool←⎕C Args._1
      :If ~(⊂tool)∊'all' 'cider' 'tatin'
          →0⊣r←'Tool must be one of all|cider|tatin'
      :EndIf
      target←USERDIR,'/SessionExtensions/CiderTatin'  ⍝ Cider & Tatin
      target2←USERDIR,'/StartupSession/CiderTatin'    ⍝ For the associated `Run` function
     
      :Select Cmd
      :Case 'Activate'
          :If Args.reset
              :Select tool
              :Case 'cider'
                  3 ⎕NDELETE target,'/Cider'
              :Case 'tatin'
                  3 ⎕NDELETE target,'/Tatin'
                  3 ⎕NDELETE target2,'/Run.apl'
              :Case 'all'
                  3 ⎕NDELETE target
                  3 ⎕NDELETE target2,'/Run.apl'
              :EndSelect
          :EndIf
          :If 'all'≡tool
          :AndIf ⎕NEXISTS target
              →0⊣r←'Folder "',target,'" exists, please specify -reset'
          :EndIf
          :If 'tatin'≡tool
          :AndIf ⎕NEXISTS target,'/Tatin'
              →0⊣r←'Folder "',target,'/Tatin" exists, please specify -reset'
          :EndIf
          :If 'cider'≡tool
          :AndIf ⎕NEXISTS target,'/Cider'
              →0⊣r←'Folder "',target,'/Cider" exists, please specify -reset'
          :EndIf
          source←''
          folder←DYALOG,'/Experimental/CiderTatin/'
          :If (⊂tool)∊'tatin' 'all'
              source,←⊂folder,'Tatin'
          :EndIf
          :If (⊂tool)∊'cider' 'all'
              source,←⊂folder,'Cider'
          :EndIf
          ⍝ Tatin and Run are always copied, Cider only if "tool" is "cider" or "all"
          :If ~m←⎕NEXISTS src←folder
          :OrIf ~∧/m←⎕NEXISTS src←source
              →0⊣r←'Tool source not found: ',⍕(~m)/src
          :EndIf
     
          3 ⎕MKDIR target
          target ⎕NCOPY source
     
          source←folder,'Run.aplf'
          :If ~⎕NEXISTS target2
              3 ⎕MKDIR target2
              target2 ⎕NCOPY source
          :EndIf

          sep←(1+'Win'≡OS)⊃':;'
          cmddir←GetSaltsCmdDir sep
          cmddir,←⊂{(-'/'=¯1↑⍵)↓⍵}'/'@(⍸'\'=target)⊣target
          cmddir←∪cmddir
          cmddir←1↓⊃,/sep,¨cmddir
          z←⎕SE.SALT.Set'cmddir ',cmddir,' -permanent'
          r←'cmddir set to: ',,⎕SE.SALT.Set'cmddir'
          r←↑'Restart APL to complete activation.'r
     
      :Case 'Deactivate'
          :If ~⎕NEXISTS target
              →0⊣r←'Neither Tatin nor Cider are activated'
          :EndIf
          :If ~(⊂tool)∊'all' 'cider'
          :AndIf ⎕NEXISTS target,'/Cider'
              →0⊣r←'Tool name must be either "all" or "cider" to deactivate'
          :EndIf
     
          :If 'cider'≡tool
              target,←'/Cider'
          :EndIf
          s←≢r←'"',target,'" '
          :If ⎕NEXISTS target
              3 ⎕NDELETE target
              r,←'deleted, '
          :EndIf
          :If 'all'≡tool
          :OrIf (~⎕NEXISTS target)∧'tatin'≡tool
              :If ⎕NEXISTS target2
                  3 ⎕NDELETE target2
              :EndIf
          :EndIf
          :If ~⎕NEXISTS USERDIR,'/SessionExtensions/CiderTatin'
          :AndIf 1∊target∊z←⎕SE.SALT.Set'cmddir'
              z←⎕SE.SALT.Set'cmddir ~"',target,'" -permanent'
              r,←'removed from SALT cmddir, '
          :EndIf
          :If s=≢r
              r←'No change required - not active.'
          :Else
              r←↑'Restart APL to complete deactivation.'((¯2↓r),'.')
          :EndIf
     
      :Else
          r←'Internal error, command not found: ',Cmd
      :EndSelect
    ∇

    ∇ cmddir←GetSaltsCmdDir sep
      cmddir←⎕SE.SALT.Settings'cmddir'
      (('\'=cmddir)/cmddir)←'/'
      cmddir←sep(≠⊆⊢)cmddir
      cmddir←{(-'/'=¯1↑⍵)↓⍵}¨cmddir
    ∇

:EndNamespace
